"""
    Buffer Widget.
    Decided that converting from IRC encoding should happen separately.
    20000805: * Woo, very quickly got a working buffer. Now for the IRC encoding
                convertor.
              * Maybe proper word-wrap won't be so hard to do?
         
"""

from Widget import *
from NoDebug import *
from string import split,atoi
import re
import _curses

_split_re=re.compile('(\002|\017|\037|\026|\003(?:[0-9]{1,2}(?:,[0-9]{1,2})?)?)')  # ^B, ^O, ^_, ^V, ^C
_colour_map=(15,0,1,10,12,4,5,6,14,2,3,11,9,13,8,7) # resemble default mIRC-colours
#_colour_map=(0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15)

"""
Color code list
 0 white
 1 black
 2 blue     (navy)
 3 green
 4 red
 5 brown    (maroon)
 6 purple
 7 orange   (olive)
 8 yellow
 9 lt.green (lime)
 10 teal    (a kinda green/blue cyan)
 11 lt.cyan (cyan ?) (aqua)
 12 lt.blue (royal)
 13 pink    (light purple) (fuchsia)
 14 grey
 15 lt.grey (silver)
"""

def irc_convert(text,default_colour):
    global _split_re,_colour_map

    result=[]
    colour=default_colour
    stext=_split_re.split(text)

#    debugmsg(repr(stext))

    while stext:
        s=stext[0]                       # get first element (text)
        if len(s):                       # is there text?
            result=result+[colour,stext[0]] # append it
#        debugmsg('result now %s'%result)
#        debugmsg('stext now %s'%stext)
        if len(stext)<2:
            break
        fmt=stext[1]                     # get next element (format string)
        if fmt=='\002':                  # ^B
            colour=colour^0x08           # toggle bright foreground
        elif fmt=='\037':                # ^_
            colour=colour^0x80           # toggle bright background (which may or may be blinking)
        elif fmt=='\026':                # ^V
            colour=(colour>>4)+((colour&15)<<4) # swap fg/bg colours
        elif fmt=='\017':                # ^O
            colour=default_colour        # reset to default
        elif fmt[0]=='\003':             # ^C
            fmt=split(fmt[1:],',')
            if not fmt[0]:               # just a ^C?
                colour=default_colour    # reset to default
            else:
                colour=(colour&0xF0)+_colour_map[atoi(fmt[0])&0x0F] # foreground
                if len(fmt)>1:
                    colour=(colour&0x0F)+_colour_map[atoi(fmt[1])&0x0F]*0x10 # background
        stext=stext[2:]

    return result

class BufferWidget(Widget):
    """
    This is your typical scrollable static text buffer (with colour formatting)
    fmtecho() does most of the work

    Output to the scroll buffer pad wraps at the bottom, rather than cause
        scrolling (which for a thousand line pad wouldn't be very fast. :)

    The raw_buf array behaves more conventionally, however. It contains the
        same number of lines but some of these will be wasted if any of them
        convert to more than one formatted line.

    A line of raw, unformatted data is stored as a tuple, alternating between
        colour changes and text (in that order).

    Note that the text colour is retained until the end of the line. If you
        want the space between the end of text to eol then append
        [thecolouryouwant,''] to the text.
    """
    scroll_len=2000                      # number of formatted lines to store

    def __init__(self,window,pos):
        Widget.__init__(self,window,pos)
        self.raw_buf=[]                  # buffer of raw (unformatted) lines
        self.scroll=0                    # number (formatted) lines up from bottom
        self.scrolled=0                  # number of (formatted) lines scrolled back
        self.init()

    def init(self):
        # pad for storing formatted text scrollback
        # This widget does not remember text properly across resizes, yet!
        self.fmt_pad=_curses.newpad(BufferWidget.scroll_len,self.rect.wid()-1 +1)
        self.fmt_pad.bkgd(self.window.root.attrmap[0x34]+ord('.'))
        self.set_colour(0xB1)
#        self.fmt_pad.move(BufferWidget.scroll_len-1,0)
        for y in range(BufferWidget.scroll_len):
            self.fmt_pad.addstr(y,0,'Line %03d'%y)
        self.fmt_pos=0                   # row in buf_pad of next unused line (well, last (top) used line really)

        #
        self.buf_pad=self.window.pad.subpad(self.rect.hgt(),self.rect.wid()-1 +1,self.rect.y1(),self.rect.x1())
#        self.buf_pad.bkgd(self.window.root.attrmap[0x45]+ord('*'))

        # widget's scroll bar
        self.sb_pad=self.window.pad.subpad(self.rect.hgt(),1,self.rect.y1(),self.rect.x2()-1)
#        self.sb_pad.bkgd(self.window.root.attrmap[0x2F]+ord('|'))
        self.draw()

    def draw(self):
        start=(self.fmt_pos-self.scrolled)%BufferWidget.scroll_len-self.rect.hgt()
        if start<0: # if we have to worry about wrap
#            debugmsg('  wrap '+str((BufferWidget.scroll_len+start,0,0,0,-start,self.rect.wid()-2,0,)))
            _curses.copywin(self.fmt_pad,self.buf_pad,BufferWidget.scroll_len+start,0,0,0,-start-1,self.rect.wid()-2 +1,0)
            _curses.copywin(self.fmt_pad,self.buf_pad,0,0,-start,0,self.rect.hgt()-1,self.rect.wid()-2 +1,0)
        else:
#            debugmsg('nowrap '+str((start,0,0,0,self.rect.hgt()-1,self.rect.wid()-2,0,)))
            _curses.copywin(self.fmt_pad,self.buf_pad,start,0,0,0,self.rect.hgt()-1,self.rect.wid()-2 +1,0)
        self.sb_pad.bkgd(self.window.root.attrmap[0x2F]+ord(' '))
        self.sb_pad.addch(self.rect.hgt()-2,0,'X',self.window.root.attrmap[0x5F])

    def set_scrolled(self,amount):
        self.scrolled=amount
        self.draw()

    def set_colour(self,col):
        self.colour=col

    def fprint(self,ftext):
        # ftext must be of even length!
        if len(self.raw_buf)>BufferWidget.scroll_len:
            del self.raw_buf[0]
        self.raw_buf.append(ftext)

        attrmap=self.window.root.attrmap # shortcut
        wrapx=self.rect.wid()-2 +1
        x=0
        self.fmt_pad.move(self.fmt_pos,x)
        debugmsg('fprint(%s)'%(ftext,))
#        for colour,text in ftext:        # bah
        while ftext:
            colour,text=ftext[:2]
            ftext=ftext[2:]
#            self.fmt_pad.bkgdset(attrmap[colour]+ord('-'))
            self.fmt_pad.bkgdset(attrmap[colour])
            x=x+len(text)                # get new x
            while x>wrapx:
                x=x-wrapx                # x=number of chars spilled onto next line
                self.fmt_pad.addstr(text[:-x],attrmap[colour]) # all but last x chars on this line
                self.fmt_pad.clrtoeol()
                self.fmt_pos=(self.fmt_pos+1)%BufferWidget.scroll_len # go to next line

                self.fmt_pad.move(self.fmt_pos,0)
                text=text[-x:]           # text=last x chars of text
            self.fmt_pad.addstr(text,attrmap[colour]) # all but last x chars on this line

        self.fmt_pad.clrtoeol()
        self.fmt_pos=(self.fmt_pos+1)%BufferWidget.scroll_len # go to next line
        self.draw()

    def ircprint(self,text):
        self.fprint(irc_convert(text,self.colour))

    def oldfmtecho(self,line):
        """ Output a line of stuff.

        This can be quite slow if there are lots of colours in the text

        To do:
            add scroll-bar handling (should probably implement scrolling first, though ;-)
            add ANSI colour parsing
            add proper word wrap
            Reformat control chars (or persuade ncurses to not caret-ify them)
        """
        global attrmap

        colour=self.colour
        x=0                              # xpos of last char on line, before wrap
        wrapwid=self.pos.wid()-2         # -1 for scrollbar, -1 coz you can't write to last column in a pad
        self.buf_pad.bkgdset(attrmap[colour]+255)  # this is required (blanks to eol before \n)
        self.buf_pad.addstr('\n',attrmap[colour])
        self.buf_pad.bkgdset(attrmap[0x07]+255)  # this is required (colour is the allegedly unchangable COLOR_PAIR(0), char is the blanks inside text)
#        self.buf_pad.bkgdset(attrmap[colour]&~curses.A_BLINK)  # this is required (colour is the allegedly unchangable COLOR_PAIR(0), char is the blanks inside text)
                                         # It enables non-bright foregrounds when bright backgrounds (BLINK) are used. Don't ask.
#        debugmsg('fmtecho(\''+line+'\')')
        line=BufferWidget.format_re.split(line)  # split line into alternating text and formatting sequences
        while 1:
            s=line[0]
#            debugmsg('    \''+s+'\' x='+str(x)+' len='+str(len(s)))
            x=x+len(s)
            while x>=wrapwid:            # will this extend past end of line?
#                self.buf_pad.addstr('x='+str(x)+' wid='+str(self.pos.wid()))
#                debugmsg('        wrap')
                self.buf_pad.addstr(s[:wrapwid-x],attrmap[colour])  # note that wrapwid-x is negative
                self.buf_pad.clrtoeol()
                self.buf_pad.bkgdset(attrmap[colour]+255)  # this is required (blanks to eol before \n)
                self.buf_pad.addstr('\n',attrmap[colour])
                self.buf_pad.bkgdset(attrmap[0x07]+255)  # this is required (colour is the allegedly unchangable COLOR_PAIR(0), char is the blanks inside text)
                s=s[self.pos.wid()-2-x:]
                x=x-(self.pos.wid()-2)
#            debugmsg('        no (more) wrap')
            self.buf_pad.addstr(s,attrmap[colour])
            del line[0]
            if not line:
                break
            code=line[0]
            if code=='\002':             # ^B
                colour=colour^0x08
            elif code=='\037':           # ^_
                colour=colour^0x80
            elif code=='\026':           # ^V
                colour=(colour>>4)+((colour&15)<<4)
            elif code[0]=='\003':        # ^C
                code=string.split(code[1:],',')
                if not code[0]:          # just a ^C?
                    colour=self.colour
                else:
                    colour=(colour&0xF0)+(string.atoi(code[0])&0x0F)
                    if len(code)>1:
                        colour=(colour&0x0F)+(string.atoi(code[1])&0x0F)*0x10
            elif code[0]=='\033':        # CSI sequences.
               pass                      # ANSI sequences to be implemented later
            del line[0]
        self.buf_pad.bkgdset(attrmap[self.colour]+ord('=')*0+255) # colour and char of space to eol (until \n is output)
        if not self.scroll:
            print (BufferWidget.scroll_len-self.pos.hgt()-2,0)+self.buf_pos.hwyx()
            apply(self.buf_pad.noutrefresh,tuple((BufferWidget.scroll_len-self.pos.hgt()-2,0)+self.buf_pos.hwyx()))
        self.buf_pad.clrtoeol()

    def echo(self,line):
        self.buff.append(line)
#        if len(self.buff)>=BufferWidget.buff_len:          # prevent buffer from growing unmanagably large
#            del self.buff[0]

        self.fmtecho(line)
        self.fresh()

